import itertools
from DescriptorElement import DescriptorElement
from DistanceMatrix import DistanceMatrix
from rdkit import Chem

__author__ = 'modlab'


class MolChargeAutocorrelation(DescriptorElement):
    description = '3D Autocorrelation function for whole molecule'

    def __init__(self, **options):
        self.charge_type = options.pop('charge_type')
        self.r_min = options.pop('rmin', 1.0)
        self.r_max = options.pop('rmax', 11.0)
        self.step_size = options.pop('step_size', 0.2)
        self.range = int(self.r_max - self.r_min)
        self.no_h = options.pop('no_h', False)
        self.united_atoms = options.pop('united_atoms', False)
        if self.no_h and self.united_atoms:
            raise Exception('Both options set: united atoms and no hydrogen. This is invalid.')
        # The range_values are starting from 0 and are the range divided into bins of step size 'step_size'
        self.range_values = [self.step_size * step for step in range(int(self.range / self.step_size))]
        super(MolChargeAutocorrelation, self).__init__(**options)

    def calculate_elements(self, molecule):
        """ Overridden base method that is interface to calculate and return the descriptor's x-vector

        Casts charges as properties on atoms that they can be manipulated further.
        Calls specific molecule manipulation method if special H-atom processing is specified.

        :param molecule: RDkit molecule
        :return: x-vector of descriptor
        :rtype: list
        """
        # cast charges on atoms that atoms can be manipulated
        # Note: I checked that atoms keep their properties when mol is copied (i.e. with Chem.RemoveHs() or Chem.Mol() )
        # in the RDkit implementation. Otherwise this could give a really nasty bug !!!
        charges = [charge for charge in molecule.GetProp(self.charge_type).split()]
        for atom in molecule.GetAtoms():
            atom.SetProp('Charge', charges[atom.GetIdx()])
        # Manipulate molecule according H treatment flag that was specified in the constructor
        if self.no_h:
            molecule = self._remove_h_atoms(molecule)
        elif self.united_atoms:
            molecule = self._unite_atoms(molecule)
        # Calculate autocorrelation vector for molecule
        ac_vec = self._calculate_elements(molecule)
        return ac_vec

    def _calculate_elements(self, molecule):
        """ Function that calculated the autocorrelation descriptor for molecule.

        Current implementation: If H atoms should be ignored or an united atom approach is chosen, the
        molecule is manipulated accordingly and then handed over this function.

        :param molecule: RDkit molecule
        :return: Autocorrelation vector for molecule
        :rtype: list
        """
        atoms = molecule.GetAtoms()
        self.charges = [float(atom.GetProp('Charge')) for atom in atoms]
        dm = DistanceMatrix(molecule)
        ac_vector = []
        for val in self.range_values:
            ac_vector.append(self._autocorrelation(self.r_min + val, self.r_min + val + self.step_size, atoms,
                                                   dm.distance_matrix))
        return ac_vector

    def _autocorrelation(self, r_min, r_max, atoms, dm):
        """ Calculates the accumulated crosscorrelation for atoms in bin spezified by r_min and r_max
        :param r_min: Lower bound (distance) for bin
        :param r_max: Upper bound (distance) for bin
        :param atoms: list of all atoms in molecule
        :param dm: distance matrix (n_atoms x n_atoms); needed to calculate if atom pair is in bin
        :return: Autocorrelation value for bin
        :rtype: float
        """
        assert len(atoms) == len(self.charges)
        atom_indices = range(len(atoms))
        accum = 0.0
        # Itertools combinations does not count twice. I.e. combination i,j is not repeated as j,i
        for i, j in itertools.combinations(atom_indices, 2):  # Note: does not correlate with itself ... !!
            if r_min < dm[i,j] <= r_max:
                # accum += atoms[i].GetProp(self.charge_type) * atoms[j].GetProp(self.charge_type)
                accum += self.charges[i] * self.charges[j]
        return accum

    def _remove_h_atoms(self, molecule):
        """ Removes H atoms from molecule
        :param molecule: RDkit molecule from which H atoms should be removed
        :return: RDkit molecule with removed H atoms
        """
        molecule = Chem.RemoveHs(molecule)
        return molecule

    def _unite_atoms(self, molecule):
        # TODO print a molecule with united charge and compare with non-united charge to finally veryfiy that this works
        # TODO I checked for two atoms in one example. but if visual inspection is possible this should be a better verification
        """ Sums H-atom charges into bound heavy atoms and removes H atoms from molecule
        :param molecule: RDkit molecule for which atoms should be united
        :return: RDkit molecule with H atom charges summed into heavy atoms and H atoms removed
        """
        # Make initial list of charges
        charges = [float(charge) for charge in molecule.GetProp(self.charge_type).split()]
        new_charges = [0 for _ in charges]
        # Search for heavy atoms and bound H's and sum together charges
        for atom in molecule.GetAtoms():
            if atom.GetAtomicNum() is not 1:
                a_idx = atom.GetIdx()
                charge = charges[a_idx]
                for neighbour in atom.GetNeighbors():
                    if neighbour.GetAtomicNum() is 1:
                        charge += charges[neighbour.GetIdx()]
                new_charges[a_idx] = charge
        # DEBUG all H atom charges should be zero! ---> could be moved into a test ... :-) :-) testing does eock for these purposes
        for atom in molecule.GetAtoms():
            if atom.GetAtomicNum() is 1:
                assert new_charges[atom.GetIdx()] == 0
        # Assign new charges to atoms in the molecule instance, so that Hs can be removed. This procedure should
        # assure that atoms bear correct charges even if the order is messed up upon H removal
        tmp_mol = Chem.Mol(molecule)
        for atom in tmp_mol.GetAtoms():
            a_idx = atom.GetIdx()
            atom.SetProp('Charge', str(new_charges[a_idx]))
        tmp_mol = Chem.RemoveHs(tmp_mol)
        return tmp_mol