from rdkit import Chem
from rdkit.Chem import Descriptors
from DescriptorElement import DescriptorElement

__author__ = 'modlab'


class LabuteSurface(DescriptorElement):
    description = "Labute Surface area of the atom"

    def __init__(self, **options):
        super(LabuteSurface, self).__init__(**options)

    def calculate_elements(self, atom):
        parent_mol = atom.GetOwningMol()
        try:
            surf_areas = parent_mol._labuteContribs[1:]
            tot_surf = parent_mol._labute_tot
        except AttributeError:
            tot_surf = parent_mol._labute_tot = Chem.MolSurf.pyLabuteASA(parent_mol)
            surf_areas = parent_mol._labuteContribs[1:]
        return [surf_areas[atom.GetIdx()], surf_areas[atom.GetIdx()] / tot_surf]


class MolLogP(DescriptorElement):
    description = "RDKit MolLogP descriptor"

    def __init__(self, **options):
        super(MolLogP, self).__init__(**options)

    def calculate_elements(self, atom):
        parent_mol = atom.GetOwningMol()
        try:
            mollogp = parent_mol._mol_logp
        except AttributeError:
            mollogp = parent_mol._mol_logp = Descriptors.MolLogP(parent_mol)
        return [mollogp]


class MolMR(DescriptorElement):
    description = 'RDKIT MolMR descriptor'

    def __init__(self, **options):
        super(MolMR, self).__init__(**options)

    def calculate_elements(self, atom):
        parent_mol = atom.GetOwningMol()
        try:
            mol_mr = parent_mol._mol_mr
        except AttributeError:
            mol_mr = parent_mol._mol_mr = Descriptors.MolMR(parent_mol)
        return [mol_mr]


class Hybridization(DescriptorElement):
    description = "RDKit Hybridization descriptor or Aromatic"

    def __init__(self, **options):
        super(Hybridization, self).__init__(**options)

    def calculate_elements(self, atom):
        hybrid = str(atom.GetHybridization())
        if 'SP3' in hybrid:
            hybrid = 3
        elif 'SP2' in hybrid:
            hybrid = 2
        else:
            hybrid = 1
        if atom.GetIsAromatic():
            hybrid = 0
        return [hybrid]


class TPSA(DescriptorElement):
    description = "RDKit TPSA descriptor"

    def __init__(self, **options):
        super(TPSA, self).__init__(**options)

    def calculate_elements(self, atom):
        parent_mol = atom.GetOwningMol()
        try:
            tpsa_all = parent_mol._tpsa_total
            tpsa_contribs = parent_mol._tpsa_contribs
        except AttributeError:
            tpsa_all = parent_mol._tpsa_total = Descriptors.TPSA(parent_mol)
            tpsa_contribs = parent_mol._tpsa_contribs = Chem.MolSurf._pyTPSAContribs(parent_mol)
        tpsa_atom = tpsa_contribs[atom.GetIdx()]
        if tpsa_all > 0.1:
            return [tpsa_atom, tpsa_atom / tpsa_all]
        else:
            return [0., 0.]


class ExactMolWt(DescriptorElement):
    description = "RDKit ExactMolWt descriptor"

    def __init__(self, **options):
        super(ExactMolWt, self).__init__(**options)

    def calculate_elements(self, atom):
        parent_mol = atom.GetOwningMol()
        # TODO save val to mol to make recalc not required
        return [Descriptors.ExactMolWt(parent_mol)]


class Fii(DescriptorElement):
    """
    Returns the diagonal element of the Fock Matrix in the basis of Foster-Boys LMOs, contained in the sdf
    files as molecular property
    """
    description = "F_ii of LP/pi localized orbital in eV"

    def __init__(self, **options):
        super(Fii, self).__init__(**options)

    def calculate_elements(self, atom, *other_objects):
        """
        Calculate descriptor elements. Returns descriptor as list.

        :param atom: RDKit atom to be characterized.
        :param other_objects: list of other objects. First element needs to be the molecule.
        :return: list of the six descriptor elements.
        """
        parent_mol = atom.GetOwningMol()
        query_a_idx = atom.GetIdx() + 1
        # list of LMO energy entries:
        lmo_info = parent_mol.GetProp('LP/Pi Boys F_ii elements (eV)').split('\n')
        # return LMO energy on query atom for descriptor
        lmoener = 999
        for line in lmo_info:
            lmo_a_idx = int(line.split()[-1])
            if lmo_a_idx == query_a_idx:
                lmoener = float(line[0:6])
                return [lmoener]

        if lmoener == 999:
            print "Warning: Not all HBA atoms have LP/Pi F_ii values."
            return [0.0]

        #if parent_mol.GetProp('LP/Pi Boys F_ii elements (eV)').__contains__('-1'):
        #    return [float(parent_mol.GetProp('LP/Pi Boys F_ii elements (eV)').split()[0].replace(',',''))]
        #else:
        #    return 0.0
