from DescriptorElement import DescriptorElement
from collections import Counter

__author__ = 'modlab'


# -------------------------------- helper functions ---------------------------------------------------
def determine_symbol(atom):
    symbol = atom.GetSymbol()
    # determine if aromatic
    if atom.GetIsAromatic():
        symbol += '_ar'
    # otherwise set hybridization
    else:
        symbol += '_' + str(atom.GetHybridization())
    return symbol


def get_set_of_som_symbols(dc):
    """ NOTE: The descriptorCreator needs the SoM property on atoms !!!!
    :param dc: DescriptorCreator
    :return: set containing all symbols that are SoM
    """
    som_symbol = []
    symbol_set = set()
    for molecule in dc.suppl:
        for atom in molecule.GetAtoms():
            assert atom.HasProp('som'), 'Not all atoms in the DescriptorCreator.suppl do som property'
            if atom.GetProp('som') == '1':
                atom_tuple = (atom.GetSymbol(), atom.GetHybridization(), atom.GetIsAromatic())
                som_symbol.append(atom_tuple)
                symbol = determine_symbol(atom)
                symbol_set.add(symbol)
    # for symbol in sorted(symbol_set):
    #     print symbol
    return symbol_set


def cast_sdf_mol_props_to_atoms(supplier):
    """ Assign atomic properties stored as molecular property to atom instances.

    A lot of atomic properties are stored as list of number as molecular properties in an SD File.
    This method casts them on atoms, such that later functions only needs to loop over atoms.

    :param supplier: RDKit supplier
    :return: None
    """
    for molecule in supplier:
        atoms = molecule.GetAtoms()
        # Set Atom property if it is SOM:
        som_indices = get_som_indices(molecule)
        for atom in atoms:
            if atom.GetIdx() in som_indices:
                atom.SetProp('som', '1')
            else:
                atom.SetProp('som', '0')


def get_som_indices(molecule):
    """ Returns SoM indices for an that are stored as proerty in an RDKit molecule
    :param molecule: RDKit molecule
    :return: list of som indices
    :rtype: list
    """
    # Note that indices are already here converted to python standard, i.e. starting with 0 ...
    prop_names = molecule.GetPropNames()
    som_prop_names = [name for name in prop_names if 'SOM' in name]
    som_indices = set()
    for spn in som_prop_names:
        indices = molecule.GetProp(spn)
        for index in indices.split():
            som_indices.add(int(index)-1)
    return som_indices


# -------------------------------- classes ---------------------------------------------------
class SomRatioCalculator(object):
    def __init__(self, som_container):
        """
        :param som_container: Either iterable (mol supplier) or molecule
        :return:
        """
        self.som_symbols = []
        self.non_som_symbols = []
        # create list of SOM and non-SOM atom types
        try:
            for molecule in som_container:
                self._extract_som(molecule)
        except TypeError:
            self._extract_som(som_container)
        # create Counter dictionaries from which the ratios can be calculated
        self.som_counter = Counter(self.som_symbols)
        self.non_som_counter = Counter(self.non_som_symbols)

    def _extract_som(self, molecule):
        """ Extracts SoMs from molecule and adds them to state
        :param molecule: RDKit molecule
        :return: None
        """
        for atom in molecule.GetAtoms():
            symbol = determine_symbol(atom)
            if atom.GetProp('som') == '1':
                self.som_symbols.append(symbol)
            else:
                self.non_som_symbols.append(symbol)

    def som_ratios(self):
        """ Returns a dictionary containing som string and their ration SOM/non-SOM in this molecule
        :return: dict with key=SOM string and value=ratio SOM/non-SOM
        :rtype: dict
        """
        d = dict()
        for key in self.som_counter:
            try:
                d[key] = float(self.som_counter[key]) / float(self.non_som_counter[key])
            except ZeroDivisionError:
                # print 'Zero Division for following atom type: {}'.format(key)
                d[key] = float(self.som_counter[key])
        return d

    def ratio_for_atom_type(self, atomtype):
        """
        :param atomtype: atom type for which the SoM ratio should be determined
        :type atomtype: str
        :return: ratio for atom type of SoM/ (SoM + non-SoM) of stored set
        :rtype: float
        """
        try:
            ratio = float(self.som_counter[atomtype]) / \
                    (float(self.som_counter[atomtype]) + float(self.non_som_counter[atomtype]))
        except ZeroDivisionError:
            ratio = 0
        return ratio


class MolCollectionSomRatios(object):
    # NOTE: DEPRECATED !!!!!
    def __init__(self, mol_suppl):
        self.som_symbols = []
        self.non_som_symbols = []
        # create list of SOM and non-SOM atomtypes
        for molecule in mol_suppl:
            for atom in molecule.GetAtoms():
                symbol = determine_symbol(atom)
                if atom.GetProp('som') == '1':
                    self.som_symbols.append(symbol)
                else:
                    self.non_som_symbols.append(symbol)
        # create Counter dictionaries from which the ratios can be calculated
        self.som_counter = Counter(self.som_symbols)
        self.non_som_counter = Counter(self.non_som_symbols)

    def som_ratios(self):
        """ Returns a dictitonary containing som string and their ration SOM/non-SOM in this molecule
        :return: dict with key=SOM string and value=ratio SOM/non-SOM
        :rtype: dict
        """
        d = dict()
        for key in self.som_counter:
            try:
                d[key] = float(self.som_counter[key]) / \
                         (float(self.non_som_counter[key]) + float(self.som_counter[key]))
            except ZeroDivisionError as e:
                print e
                raise e
        return d


class SomDistribution(DescriptorElement):
    description = 'Descriptor that stores Atomtypes and SOM as histogramm'

    def __init__(self, **options):
        # This is yet a bit difficult from a documentation point of view but:
        # The following option is absolutely required and must be handed over
        # So that all descriptors are of the same length
        self.som_types = options.pop('som_types')
        super(SomDistribution, self).__init__(**options)

    def calculate_elements(self, molecule):
        descriptor_x = []
        mol_som_ratios = SomRatioCalculator(molecule)
        som_ratios = mol_som_ratios.som_ratios()
        for key in sorted(self.som_types):
            try:
                descriptor_x.append(som_ratios[key])
            except KeyError:
                descriptor_x.append(0.)
        return descriptor_x



